""" global constants """
# for shortest path calculation
MAX_LABEL_COST = 2147483647
EPSILON = 0.00001
# for column generation
MIN_COL_VOL = 0.1
# for accessibility evaluation
MIN_TIME_BUDGET = 10
MAX_TIME_BUDGET = 240
BUDGET_TIME_INTVL = 5
# unit
MILE_TO_METER = 1609
MPH_TO_KPH = 1.609
# simulation
SECONDS_IN_MINUTE = 60
SECONDS_IN_HOUR = 3600